
// Standardne pre MS win aplikacie
#include <windows.h>

// GLEW kniznica. pre spravu novsich rozsireni
#include "glew.h"
// GL Utility Toolkit (GLUT), v ramci toho sa includuje aj openGL hlavickovy subor
#include "glut.h"
// OpenIL kniznica pre nacitanie obrazkov
#include "il.h"
// kniznica pre TrueType fonty
#include <ft2build.h>
#include <freetype/freetype.h>
#include <freetype/ftglyph.h>
#include <freetype/ftoutln.h>
#include <freetype/fttrigon.h>

#include <stdio.h>
#include <math.h>

#define PI 3.14159

// Globalne premenne
GLboolean g_isAnim = true;
GLboolean g_isSmooth = true;
GLboolean g_isTexture = false;
float sun_rotation = 0.0f;

// Globalne premenne pre prezeranie sceny
GLdouble vzd, old_vzd, fi, xi, old_fi, old_xi;
GLint left_mouse_down_x, left_mouse_down_y;
GLint right_mouse_down_y;
GLboolean right_down = false, left_down = false;

GLuint g_window_height, g_window_width;
GLuint g_2DtextWGL = 0, g_3DtextWGL = 0;

// id textur
GLuint tex_sun, tex_earth, tex_jupiter, tex_phobos;

// id shaderov
GLuint shaders_pervertex_phong, shaders_perpixel_phong;

// id VBO
GLuint g_uiVBOSphereCoords, g_uiVBOSphereTexCoords, g_uiVBOSphereNormals, g_uiVBOSphereIndices;
GLuint g_uiSphereNumIndices;

// funkcia pre nacitanie textoveho suboru do pola znakov
GLchar* readShaderFile(const char *fileName)
{
    FILE* file = fopen(fileName, "r" );
    if( file == NULL )
    {
        printf("Cannot open file %s. \n", fileName);
		return NULL;
    }

    // ziskaj velkost suboru v bajtoch
    fseek(file, 0L, SEEK_END);
    long fileSize = ftell(file);
    fseek(file, 0L, SEEK_SET);

    // nacitaj data zo suboru
    GLchar* buffer = new GLchar[fileSize + 1];
	int bytes = fread(buffer, 1, fileSize, file);
    buffer[bytes] = 0;
	fclose(file);

	return buffer;
}

// nacitanie zdrojoveho suboru shadera a ulozenie 
GLuint loadShader(GLenum shaderType, const char *fileName)
{
    GLchar *shaderSource = readShaderFile(fileName);
    if (shaderSource == NULL)
        return 0;

    GLuint shader = glCreateShader(shaderType);
    glShaderSource(shader, 1, (const GLchar**)&shaderSource, NULL);
    glCompileShader(shader);
    delete shaderSource;

    GLint status;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
    if (status == GL_FALSE)  
    {
        GLchar log[1024];
        GLsizei length;
        glGetShaderInfoLog(shader, 1024, &length, log);
        printf("Failed compiling shader %s \n", fileName);
        printf("   Compilation error: %s \n", log);
        glDeleteShader(shader);

        return 0;
    }

    printf("Shader %s compiled successfully.\n", fileName);
    return shader;
}

GLuint loadProgram(const char *fileVertexShader, const char *fileFragmentShader)
{
    // nacitanie a kompilacia shaderov
    GLuint vertexShader = loadShader(GL_VERTEX_SHADER, fileVertexShader);
    if (vertexShader == 0)
        return 0;
    GLuint fragmentShader = loadShader(GL_FRAGMENT_SHADER, fileFragmentShader);
    if (fragmentShader == 0)
    {
        glDeleteShader(vertexShader);
        return 0;
    }

    // vytvorenie program objektu a pripojenie shaderov
    GLuint programObj = glCreateProgram();
    glAttachShader(programObj, vertexShader);
    glAttachShader(programObj, fragmentShader);
    glLinkProgram(programObj);

    // zisti uspesnost zlinkovania shaderov
    GLint status;
    glGetProgramiv(programObj, GL_LINK_STATUS, &status);
    if (status == GL_FALSE)  
    {
        GLchar log[1024];
        GLsizei length;
        glGetProgramInfoLog(programObj, 1024, &length, log);
        printf("Failed linking shaders with error %s\n", log);

        glDeleteProgram(programObj);
        glDeleteShader(fragmentShader);
        glDeleteShader(vertexShader);

        return 0;
    }

    printf("Shaders linked successfully.\n");
    return programObj;
}

void initShaders()
{
    // nacitanie, kompilacia a zlinkovanie shaderov
    shaders_pervertex_phong = loadProgram("shaders/pervertex_phong.vert", "shaders/pervertex_phong.frag");
    int locationTexture = glGetUniformLocation(shaders_pervertex_phong, "color_texture");
    glUniform1i(locationTexture, 0);

    shaders_perpixel_phong = loadProgram("shaders/perpixel_phong.vert", "shaders/perpixel_phong.frag");
    locationTexture = glGetUniformLocation(shaders_perpixel_phong, "color_texture");
    glUniform1i(locationTexture, 0);

    glActiveTexture(GL_TEXTURE0);
}

// inicializacia OpenGL textur
void initTextures()
{
    // iniciailizacia OpenIL
    ilInit();

    // ************* prva textura, zem *************
    // generovanie id textury pre OpenIL
    ILuint texIL;
    ilGenImages(1, &texIL);
    ilBindImage(texIL);
    // nacitanie textury zo suboru
    bool result = ilLoadImage("earth.jpg");
    // generovanie OpenGL textur
    glGenTextures(1, &tex_earth);
    glBindTexture(GL_TEXTURE_2D, tex_earth);
    // nastavenie parametrov textury
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    // definovanie texlov
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ilGetInteger(IL_IMAGE_WIDTH), ilGetInteger(IL_IMAGE_HEIGHT), 
                 0, ilGetInteger(IL_IMAGE_FORMAT), GL_UNSIGNED_BYTE, ilGetData());
    // zamzanie OpenIL textury
    ilDeleteImages(1, &texIL);

    // ************* druha textura, slnko *************
    // generovanie id textury pre OpenIL
    ilGenImages(1, &texIL);
    ilBindImage(texIL);
    // nacitanie textury zo suboru
    result = ilLoadImage("sun.jpg");
    // generovanie OpenGL textur
    glGenTextures(1, &tex_sun);
    glBindTexture(GL_TEXTURE_2D, tex_sun);
    // nastavenie parametrov textury
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    // definovanie texlov
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ilGetInteger(IL_IMAGE_WIDTH), ilGetInteger(IL_IMAGE_HEIGHT), 
                 0, ilGetInteger(IL_IMAGE_FORMAT), GL_UNSIGNED_BYTE, ilGetData());
    // zamzanie OpenIL textury
    ilDeleteImages(1, &texIL);

    // ************* druha textura, jupiter *************
    // generovanie id textury pre OpenIL
    ilGenImages(1, &texIL);
    ilBindImage(texIL);
    // nacitanie textury zo suboru
    result = ilLoadImage("jupiter.jpg");
    // generovanie OpenGL textur
    glGenTextures(1, &tex_jupiter);
    glBindTexture(GL_TEXTURE_2D, tex_jupiter);
    // nastavenie parametrov textury
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    // definovanie texlov
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ilGetInteger(IL_IMAGE_WIDTH), ilGetInteger(IL_IMAGE_HEIGHT), 
                 0, ilGetInteger(IL_IMAGE_FORMAT), GL_UNSIGNED_BYTE, ilGetData());
    // zamzanie OpenIL textury
    ilDeleteImages(1, &texIL);

    // ************* druha textura, phobos *************
    // generovanie id textury pre OpenIL
    ilGenImages(1, &texIL);
    ilBindImage(texIL);
    // nacitanie textury zo suboru
    result = ilLoadImage("phobos.jpg");
    // generovanie OpenGL textur
    glGenTextures(1, &tex_phobos);
    glBindTexture(GL_TEXTURE_2D, tex_phobos);
    // nastavenie parametrov textury
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    // definovanie texlov
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ilGetInteger(IL_IMAGE_WIDTH), ilGetInteger(IL_IMAGE_HEIGHT), 
                 0, ilGetInteger(IL_IMAGE_FORMAT), GL_UNSIGNED_BYTE, ilGetData());
    // zamzanie OpenIL textury
    ilDeleteImages(1, &texIL);
}

void initSphereVBO(int xSlices, int ySlices)
{
    float* vertices = new float[3 * (xSlices + 1) * (ySlices + 1)];
    float* normals = new float[3 * (xSlices + 1) * (ySlices + 1)];
    float* texcoords = new float[2 * (xSlices + 1) * (ySlices + 1)];

    for (int i = 0; i <= xSlices; i++)
    {
        float u = i / float(xSlices);
        for (int j = 0; j <= ySlices; j++)
        {
            float v = j / float(ySlices);

            int index = i + j * (xSlices + 1);

            vertices[3 * index + 0] = cos(2 * u * PI) * cos((v - 0.5f) * PI);
            vertices[3 * index + 1] = sin(2 * u * PI) * cos((v - 0.5f) * PI);
            vertices[3 * index + 2] = sin((v - 0.5f) * PI);

            normals[3 * index + 0] = vertices[3 * index + 0];
            normals[3 * index + 1] = vertices[3 * index + 1];
            normals[3 * index + 2] = vertices[3 * index + 2];

            texcoords[2 * index + 0] = u;
            texcoords[2 * index + 1] = 1.0f - v;
        }
    }

    GLuint* indices = new GLuint[4 * (xSlices + 0) * (ySlices + 0)];
    int index = 0;
    for (int i = 0; i < xSlices; i++)
    {
        for (int j = 0; j < ySlices; j++)
        {
            indices[index] = i + j * (xSlices + 1);
            index++;
            indices[index] = (i + 1) + j * (xSlices + 1);
            index++;
            indices[index] = (i + 1) + (j + 1) * (xSlices + 1);
            index++;
            indices[index] = i + (j + 1) * (xSlices + 1);
            index++;
        }
    }

    glGenBuffers(1, &g_uiVBOSphereCoords);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereCoords);
    glBufferData(GL_ARRAY_BUFFER, 3 * (xSlices + 1) * (ySlices + 1) * sizeof(float), vertices, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOSphereNormals);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereNormals);
    glBufferData(GL_ARRAY_BUFFER, 3 * (xSlices + 1) * (ySlices + 1) * sizeof(float), normals, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOSphereTexCoords);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereTexCoords);
    glBufferData(GL_ARRAY_BUFFER, 2 * (xSlices + 1) * (ySlices + 1) * sizeof(float), texcoords, GL_STATIC_DRAW);

    glGenBuffers(1, &g_uiVBOSphereIndices);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, g_uiVBOSphereIndices);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 4 * (xSlices + 0) * (ySlices + 0) * sizeof(GLuint), indices, GL_STATIC_DRAW);

    g_uiSphereNumIndices = 4 * (xSlices + 0) * (ySlices + 0);

    delete [] vertices;
    delete [] normals;
    delete [] texcoords;
    delete [] indices;
}

// Open GL Inicializacia
bool init(void)
{
    // Cierne pozadie
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    
    // Nastav z-buffer testovanie
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    // nastav stencil buffer
    glClearStencil(0);

    // nastav zmiesavaciu funkciu
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // nastav vhodne mapovanie textur
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    // initializuj kameru
    fi = 0.7f; xi = 0.7f; vzd = 10.0f;

    // initializuje svetlo
    GLfloat light_ambient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
    GLfloat light_diffuse[] = { 0.9f, 0.9f, 0.9f, 1.0f };
    GLfloat light_specular[] = { 0.5f, 0.5f, 0.5f, 1.0f };
    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
    glEnable(GL_LIGHT0);

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	return true;
}

void CALLBACK beginCallback(GLenum which)
{
    glBegin(which);
}

void CALLBACK errorCallback(GLenum errorCode)
{
    printf("Tessellation Error: %s\n", gluErrorString(errorCode));
}

void CALLBACK endCallback(void)
{
    glEnd();
}

void CALLBACK vertexCallback(GLvoid *vertex)
{
    const GLdouble* pointer = (GLdouble*) vertex;
    glColor3dv(pointer+3);
    glVertex3dv(pointer);
}

void CALLBACK combineCallback(GLdouble coords[3], 
                              GLdouble *vertex_data[4],
                              GLfloat weight[4], GLdouble **dataOut )
{
    GLdouble *vertex;
    int i;
    
    vertex = (GLdouble *) malloc(6 * sizeof(GLdouble));
    
    vertex[0] = coords[0];
    vertex[1] = coords[1];
    vertex[2] = coords[2];
    for (i = 3; i < 6; i++)
        vertex[i] = weight[0] * vertex_data[0][i] 
                  + weight[1] * vertex_data[1][i]
                  + weight[2] * vertex_data[2][i] 
                  + weight[3] * vertex_data[3][i];
    *dataOut = vertex;
}

void renderTessPolygons()
{
    GLdouble quad[4][3] = {0.0, 0.5, 0.0,
                           0.5, 1.0, 0.0,
                           1.0, 0.5, 0.0,
                           0.5, 2.0, 0.0};
    GLdouble rect[4][3] = {1.5, 0.5, 0.0,
                           3.0, 0.5, 0.0,
                           3.0, 2.0, 0.0,
                           1.5, 2.0, 0.0};
    GLdouble tri[3][3] = {1.7, 0.7, 0.0,
                          2.2, 1.7, 0.0,
                          2.7, 0.7, 0.0};
    GLdouble tri2[3][3] = {2.2, 0.8, 0.0,
                           2.0, 1.0, 0.0,
                           2.4, 1.0, 0.0};
    GLdouble star[5][6] = {3.5, 0.5, 0.0, 1.0, 0.0, 1.0,
                           4.2, 2.0, 0.0, 1.0, 1.0, 0.0,
                           5.0, 0.5, 0.0, 0.0, 1.0, 1.0,
                           3.5, 1.5, 0.0, 1.0, 0.0, 0.0,
                           5.0, 1.5, 0.0, 0.0, 1.0, 0.0};
    
    // vytvor novy tesselation objekt
    GLUtesselator *tobj = gluNewTess();

    // vlastnosti teselacie
    //gluTessProperty(tobj, GLU_TESS_BOUNDARY_ONLY, GL_TRUE);
    gluTessProperty(tobj, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ODD);
    // GLU_TESS_WINDING_ODD (the default), 
    // GLU_TESS_WINDING_NONZERO, 
    // GLU_TESS_WINDING_POSITIVE, 
    // GLU_TESS_WINDING_NEGATIVE, 
    // GLU_TESS_WINDING_ABS_GEQ_TWO 

    gluTessCallback(tobj, GLU_TESS_VERTEX, (void (CALLBACK *)()) glVertex3dv);
    gluTessCallback(tobj, GLU_TESS_BEGIN, (void (CALLBACK *)()) beginCallback);
    gluTessCallback(tobj, GLU_TESS_END, (void (CALLBACK *)()) endCallback);
    gluTessCallback(tobj, GLU_TESS_ERROR, (void (CALLBACK *)()) errorCallback);
    
    glColor3f(1.0f, 0.0f, 0.0f);
    glShadeModel(GL_FLAT);    
    gluTessBeginPolygon(tobj, NULL);
      gluTessBeginContour(tobj);
        gluTessVertex(tobj, quad[0], quad[0]);
        gluTessVertex(tobj, quad[1], quad[1]);
        gluTessVertex(tobj, quad[2], quad[2]);
        gluTessVertex(tobj, quad[3], quad[3]);
      gluTessEndContour(tobj);
    gluTessEndPolygon(tobj);

    glColor3f(0.0f, 0.0f, 1.0f);
    glShadeModel(GL_FLAT);   
    gluTessBeginPolygon(tobj, NULL);
      gluTessBeginContour(tobj);
        gluTessVertex(tobj, rect[0], rect[0]);
        gluTessVertex(tobj, rect[1], rect[1]);
        gluTessVertex(tobj, rect[2], rect[2]);
        gluTessVertex(tobj, rect[3], rect[3]);
      gluTessEndContour(tobj);
      gluTessBeginContour(tobj);
        gluTessVertex(tobj, tri[0], tri[0]);
        gluTessVertex(tobj, tri[1], tri[1]);
        gluTessVertex(tobj, tri[2], tri[2]);
      gluTessEndContour(tobj);
      gluTessBeginContour(tobj);
        gluTessVertex(tobj, tri2[0], tri2[0]);
        gluTessVertex(tobj, tri2[1], tri2[1]);
        gluTessVertex(tobj, tri2[2], tri2[2]);
      gluTessEndContour(tobj);
    gluTessEndPolygon(tobj);
    
    gluTessCallback(tobj, GLU_TESS_VERTEX, (void (CALLBACK *)()) vertexCallback);
    gluTessCallback(tobj, GLU_TESS_COMBINE, (void (CALLBACK *)()) combineCallback);
    
    glShadeModel(GL_SMOOTH);    
    gluTessProperty(tobj, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_POSITIVE);
    gluTessBeginPolygon(tobj, NULL);
      gluTessBeginContour(tobj);
        gluTessVertex(tobj, star[0], star[0]);
        gluTessVertex(tobj, star[1], star[1]);
        gluTessVertex(tobj, star[2], star[2]);
        gluTessVertex(tobj, star[3], star[3]);
        gluTessVertex(tobj, star[4], star[4]);
      gluTessEndContour(tobj);
    gluTessEndPolygon(tobj);
    glEndList();

    gluDeleteTess(tobj);
}

void inti2DFontWGL(GLvoid)
{
	g_2DtextWGL = glGenLists(96);
	HFONT font = CreateFont(	-24,							// Height Of Font
						0,								// Width Of Font
						0,								// Angle Of Escapement
						0,								// Orientation Angle
						FW_BOLD,						// Font Weight
						FALSE,							// Italic
						FALSE,							// Underline
						FALSE,							// Strikeout
						ANSI_CHARSET,					// Character Set Identifier
						OUT_TT_PRECIS,					// Output Precision
						CLIP_DEFAULT_PRECIS,			// Clipping Precision
						ANTIALIASED_QUALITY,			// Output Quality
						FF_DONTCARE|DEFAULT_PITCH,		// Family And Pitch
						"Impact");					// Font Name
    HDC hDC = wglGetCurrentDC();
	HFONT oldfont = (HFONT)SelectObject(hDC, font);
	BOOL result = wglUseFontBitmaps(hDC, 32, 96, g_2DtextWGL);
    result = wglUseFontBitmaps(hDC, 32, 96, g_2DtextWGL);
	SelectObject(hDC, oldfont);
	DeleteObject(font);
}

void inti3DFontWGL(GLvoid)
{
	g_3DtextWGL = glGenLists(256);
	HFONT font = CreateFont(	-24,							// Height Of Font
						0,								// Width Of Font
						0,								// Angle Of Escapement
						0,								// Orientation Angle
						FW_BOLD,						// Font Weight
						FALSE,							// Italic
						FALSE,							// Underline
						FALSE,							// Strikeout
						ANSI_CHARSET,					// Character Set Identifier
						OUT_TT_PRECIS,					// Output Precision
						CLIP_DEFAULT_PRECIS,			// Clipping Precision
						ANTIALIASED_QUALITY,			// Output Quality
						FF_DONTCARE|DEFAULT_PITCH,		// Family And Pitch
						"Courier New");					// Font Name
    HDC hDC = wglGetCurrentDC();
	HFONT oldfont = (HFONT)SelectObject(hDC, font);
    GLYPHMETRICSFLOAT agmf[256];
	BOOL result = wglUseFontOutlines(hDC, 32, 96, g_3DtextWGL, 0.0f, 0.1f, WGL_FONT_POLYGONS, agmf);
	SelectObject(hDC, oldfont);
	DeleteObject(font);
}

GLuint g_idTextures2DFontFT[96];
struct GlyphMetrics
{
    int left;
    int top;
    int advance;
    int bitmap_width;
    int bitmap_height;
};
GlyphMetrics g_glyphs2DFontFT[96];
bool g_2DFontFTInited = false;

void init2DFontFT()
{
    FT_Library library;
	if (FT_Init_FreeType(&library)) 
		return;

	FT_Face face;
	if (FT_New_Face(library, "planetbe.ttf", 0, &face)) 
		return;

    int fontHeight = 24;
	FT_Set_Char_Size(face, 64 * fontHeight, 64 * fontHeight, 96, 96);

    glGenTextures(96, g_idTextures2DFontFT);
    for (int i = 0; i < 96; i++)
    {
        // inicializuje bitmapu pre aktualny glyf
        if (FT_Load_Glyph(face, FT_Get_Char_Index(face, i + 32), FT_LOAD_DEFAULT))
		    continue;
        FT_Glyph glyph;
        if(FT_Get_Glyph(face->glyph, &glyph))
		    continue;
        FT_Glyph_To_Bitmap(&glyph, ft_render_mode_normal, 0, 1);
        FT_BitmapGlyph bitmap_glyph = (FT_BitmapGlyph)glyph;
        FT_Bitmap& bitmap = bitmap_glyph->bitmap;

        // uloz potrebne data pre glyf
        GlyphMetrics glyphMetrics;
        glyphMetrics.top = bitmap_glyph->top;
        glyphMetrics.left = bitmap_glyph->left;
        glyphMetrics.advance = face->glyph->advance.x >> 6;
        glyphMetrics.bitmap_width = bitmap.width;
        glyphMetrics.bitmap_height = bitmap.rows;
        g_glyphs2DFontFT[i] = glyphMetrics;

        // ak nemame bitmapu pre glyf, preskoc glyf
        if (bitmap.width == 0 || bitmap.rows == 0)
            continue;

        // vypocitaj rozmery textury, mocniny 2
        int iNX = int(log((float)bitmap.width)/log(2.0)) + 1;
        int texture_width = int(pow(2.0, iNX));
        int iNY = int(log((float)bitmap.rows)/log(2.0)) + 1;
        int texture_height = int(pow(2.0, iNY));

        // rozsir bitmapu do textury
        GLubyte* texture_data = new GLubyte[ 2 * texture_width * texture_height];
        for(int j = 0; j < texture_height; j++) 
        {
		    for(int k = 0; k < texture_width; k++)
            {
                int bitmap_x = int((bitmap.width - 1) * k / float(texture_width - 1));
                int bitmap_y = int((bitmap.rows - 1) * j / float(texture_height - 1));
			    texture_data[2*(k + j*texture_width)] = texture_data[2*(k + j*texture_width)+1] = 
                    bitmap.buffer[bitmap_x + bitmap.width*bitmap_y];
            }
		}

        // nacitaj texturu do opengl
        glBindTexture(GL_TEXTURE_2D, g_idTextures2DFontFT[i]);
	    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texture_width, texture_height, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, texture_data);
        delete [] texture_data;
    }

    FT_Done_Face(face);
	FT_Done_FreeType(library);
    g_2DFontFTInited = true;
}

void render2DTextFF(float x, float y, const char *string)
{
    if (!g_2DFontFTInited)
        init2DFontFT();

    float current_x = x;
    float current_y = y;

    glEnable(GL_TEXTURE_2D);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(0, g_window_width, 0, g_window_height);
    glEnable(GL_BLEND);

    for (const char* c = string; *c != '\0'; c++) 
    {
        if (g_glyphs2DFontFT[*c - 32].bitmap_width > 0 && g_glyphs2DFontFT[*c - 32].bitmap_height > 0)
        {
            glBindTexture(GL_TEXTURE_2D, g_idTextures2DFontFT[*c - 32]);

            int bitmap_width = g_glyphs2DFontFT[*c - 32].bitmap_width;
            int bitmap_height = g_glyphs2DFontFT[*c - 32].bitmap_height;
            float offset_x = g_glyphs2DFontFT[*c - 32].left;
            float offset_y = g_glyphs2DFontFT[*c - 32].top - bitmap_height;

            glBegin(GL_QUADS);
	        glTexCoord2d(0, 0); glVertex2f(current_x + offset_x, current_y + offset_y + bitmap_height);
	        glTexCoord2d(0, 1); glVertex2f(current_x + offset_x, current_y + offset_y);
	        glTexCoord2d(1, 1); glVertex2f(current_x + offset_x + bitmap_width, current_y + offset_y);
	        glTexCoord2d(1, 0); glVertex2f(current_x + offset_x + bitmap_width, current_y + offset_y + bitmap_height);
	        glEnd();
        }

        current_x += g_glyphs2DFontFT[*c - 32].advance;
    }

    glDisable(GL_BLEND);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    glDisable(GL_TEXTURE_2D);
}

void render2DTextWGL(float x, float y, const char *string)
{
    if (g_2DtextWGL == 0)
        inti2DFontWGL();

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(0, g_window_width, 0, g_window_height);
    glRasterPos2f(x, y);
    glPushAttrib(GL_LIST_BIT);
    //glWindowPos2d(x, y);
    glListBase(g_2DtextWGL - 32);
    glCallLists(strlen(string), GL_UNSIGNED_BYTE, string);
    glPopAttrib();
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
}

void render3DTextWGL(float x, float y, float z, float scale, void *font, const char *string)
{
    if (g_3DtextWGL == 0)
        inti3DFontWGL();

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glTranslatef(x, y, z);
    glRotatef(-90, 1.f, 0.0f, 0.0f);
    glScalef(-scale, -scale, scale);
    glPushAttrib(GL_LIST_BIT);
    glListBase(g_3DtextWGL - 32);
    glCallLists(strlen(string), GL_UNSIGNED_BYTE, string);
    glPopAttrib();
    glPopMatrix();
}

void render2DTextGLUT(float x, float y, void *font, const char *string)
{
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   gluOrtho2D(0, g_window_width, 0, g_window_height);
   glRasterPos2f(x, y);
   for (const char* c = string; *c != '\0'; c++) 
   {
      glutBitmapCharacter(font, *c);
   }
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();
}

void render3DTextGLUT(float x, float y, float z, float scale, void *font, const char *string)
{
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glTranslatef(x, y, z);
    glRotatef(-90, 1.f, 0.0f, 0.0f);
    glScalef(-scale, -scale, scale);
    for (const char* c = string; *c != '\0'; c++) 
    {
        glutStrokeCharacter(font, *c);
    }
    glPopMatrix();
}

void renderSphereVBO()
{
    glEnableClientState(GL_VERTEX_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereCoords);
    glVertexPointer(3, GL_FLOAT, 0, NULL);

    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereTexCoords);
    glTexCoordPointer(2, GL_FLOAT, 0, NULL);

    glEnableClientState(GL_NORMAL_ARRAY);
    glBindBuffer(GL_ARRAY_BUFFER, g_uiVBOSphereNormals);
    glNormalPointer(GL_FLOAT, 0, NULL);

    // draw a cube using buffer with indices
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, g_uiVBOSphereIndices);
    glDrawElements(GL_QUADS, g_uiSphereNumIndices, GL_UNSIGNED_INT, NULL);

    // deactivate vertex arrays after drawing
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

// Procedura obsluhujuca vykreslovanie
void render(void)   
{
    // nastavenie interpolacie farieb v ramci polygonov
    if (g_isSmooth)
        glShadeModel(GL_SMOOTH);
    else
        glShadeModel(GL_FLAT);

    // Vymaz (vypln) obrazovku a z-buffer definovanymi hodnotami
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);	

    // Nastav modelview maticu (transformaciu) na jednotkovu maticu
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();									

    // Nastav view matice pre pohlad na svetovy suradnicovy system
    GLdouble ex = vzd*cos(fi)*cos(xi);
	GLdouble ey = vzd*sin(fi)*cos(xi);
	GLdouble ez = vzd*sin(xi);
	gluLookAt( ex, ey, ez, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f );

    renderTessPolygons();

    // vykresli suradnicove osi svetoveho systemu roznymi farbami
    glLineWidth(1);
    glBegin( GL_LINES );
    glColor3f(1.0f, 0.0f, 0.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 100.0f, 0.0f, 0.0f );
    glColor3f(0.0f, 1.0f, 0.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 0.0f, 100.0f, 0.0f );
    glColor3f(0.0f, 0.0f, 1.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 0.0f, 0.0f, 100.0f );
    glEnd();

    // vykresli slnko
    glBindTexture(GL_TEXTURE_2D, tex_sun);
    if (g_isTexture)
        glEnable(GL_TEXTURE_2D);
    glPushMatrix();
    glRotatef(sun_rotation, 0.0f, 0.0f, 1.0f);
    glTranslatef(6.0f, 0.0f, 0.0f);
    // nastav poziciu svetla do stredu slnka
    GLfloat sun_position[] = { 0.0f, 0.0f, 0.0f, 1.0f };
    glLightfv(GL_LIGHT0, GL_POSITION, sun_position);
    glScalef(0.3f, 0.3f, 0.3f);
    glColor3f(1.0f, 1.0f, 1.0f );
    renderSphereVBO();
    glPopMatrix();
    glDisable(GL_TEXTURE_2D);

    // vykresli phobos
    glUseProgram(shaders_perpixel_phong);
    GLuint tex_enabled = glGetUniformLocation(shaders_perpixel_phong, "texturing_enabled");
    if (g_isTexture)
        glUniform1i(tex_enabled, 1);
    else
        glUniform1i(tex_enabled, 0);
    glBindTexture(GL_TEXTURE_2D, tex_phobos);
    glPushMatrix();
    glTranslatef(0.0f, 0.0f, -1.0f);
    glScalef(1.0f, 1.0f, 1.0f);
    renderSphereVBO();
    glPopMatrix();
    glUseProgram(0);

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);

    glColor3f(1.0f, 0.0f, 1.0f);
    render3DTextGLUT(6, 0, 2, 0.01, GLUT_STROKE_ROMAN, "Projekt OpenGL - GLUT 3D text");

    glEnable(GL_LIGHTING);
    glColor3f(0.0f, 1.0f, 1.0f);
    render3DTextWGL(6, 0, 0, 1, GLUT_STROKE_ROMAN, "Projekt OpenGL - WGL 3D text");
    glDisable(GL_LIGHTING);

    glColor3f(1.0f, 0.0f, 0.0f);
    render2DTextGLUT(20, g_window_height - 20, GLUT_BITMAP_HELVETICA_12, "Projekt OpenGL - GLUT 2D text");

    glColor3f(0.0f, 1.0f, 0.0f);
    render2DTextWGL(20, g_window_height - 50, "Projekt OpenGL - WGL 2D text");

    glColor3f(1.0f, 1.0f, 0.0f);
    render2DTextFF(20, g_window_height - 90, "Projekt OpenGL - FreeType 2D text");

    // Zamen buffre a zobraz vykreslene objekty
    glutSwapBuffers();
}

// Procedura obsluhujuca zmenu rozmerov okna
void reshape(int w, int h)
{
    g_window_width = w;
    g_window_height = h;

    // nastvenie casti okna pre vykreslenie, tentoraz kreslime do celeho okna
	glViewport(0, 0, w, h);
    // Budeme menit projekcnu maticu (transformaciu)
	glMatrixMode(GL_PROJECTION);                        
	glLoadIdentity();                                  

	if (h == 0) 
        h = 1;
    // Chceme perspektivnu projekciu
    float ratio = (float)w/(float)h;
	gluPerspective(60, ratio, 0.1, 500.0);
    //glOrtho(-10.0f*ratio, 10.0f*ratio, -10.0f, 10.0f, 1.0, 5000.0); 
}

// Procedura obsluhujuca stlacenie klavesy
void keyboard(unsigned char key, int x, int y)
{
	switch (key) 
    {
        // Ak Esc bolo stlacene
		case 27:
            // vycitime alokovane objekty
            glDeleteBuffers(1, &g_uiVBOSphereCoords);
            glDeleteBuffers(1, &g_uiVBOSphereTexCoords);
            glDeleteBuffers(1, &g_uiVBOSphereNormals);
            glDeleteBuffers(1, &g_uiVBOSphereIndices);

            // Ukoncenie programu
			exit(0);
		    break;
		default:        
		    break;
	}
}

// Procedura obsluhujuca stlacenie specialnej klavesy
void special_keys(int a_keys, int x, int y)
{
	switch (a_keys) 
    {
        // Ak F1 bolo stlacene
		case GLUT_KEY_F1:
            g_isAnim = !g_isAnim;
            glutPostRedisplay();
		    break;
        // Ak F2 bolo stlacene
        case GLUT_KEY_F2:
            g_isSmooth = !g_isSmooth;
            glutPostRedisplay();
		    break;
        // Ak F3 bolo stlacene
        case GLUT_KEY_F3:
            g_isTexture = !g_isTexture;
            glutPostRedisplay();
		    break;
		default:
		    break;
	}
}

// Procedura obsluhujuca stlacenie tlacitka mysi
void mouse_klik(int button, int state, int x, int y )
{
	switch (button) 
    {
		case GLUT_LEFT_BUTTON:
            if (state == GLUT_DOWN)
            {
                left_down = true;
			    left_mouse_down_x = x;
			    left_mouse_down_y = y;
			    old_fi = fi;
			    old_xi = xi;
                break;
            }
            else if (state == GLUT_UP)
            {
                left_down = false;
            }
		    break;
        case GLUT_RIGHT_BUTTON:
            if (state == GLUT_DOWN)
            {
                right_down = true;
			    right_mouse_down_y = y;
			    old_vzd = vzd;
                break;
            }
            else if (state == GLUT_UP)
            {
                right_down = false;
            }
		    break;
		default:
		    break;
	}

    glutPostRedisplay();
}

// Procedura obsluhujuca pohyb mysi
void mouse_move(int x, int y)
{
    if (left_down == true)
    {
        fi = old_fi + (left_mouse_down_x - x)/200.0f;
        xi = old_xi + (y - left_mouse_down_y)/200.0f;
        glutPostRedisplay();
    }

    if (right_down == true)
    {
        vzd = old_vzd + (right_mouse_down_y - y)/10.0f;
        glutPostRedisplay();
    } 

}

void timer(int id)
{
    if (g_isAnim)
    {
        sun_rotation += 0.5f;
    }

    glutPostRedisplay();
    glutTimerFunc(50, timer, 0);
}

// Zakladna funkcia spustena na zaciatku
int main(int argc, char** argv)
{
    // GLUT Inicializacia
	glutInit(&argc, argv);                              
    // Display Mod (Rgb a Double Buffering)
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
    //glutInitDisplayString("stencil>=8 rgb>=8 double depth>=24");
    // Nastavime pociatocnu poziciu a velkost okna
    glutInitWindowPosition(50, 50);
    glutInitWindowSize(800, 600);                   
    // Vyrobime okno s danym menom
    glutCreateWindow("--== Priklad 4  -  OpenGL ==--" );  

    // Zinicializujeme GLEW
    glewInit();
    // Nastavime zakladne vlastnosti OpenGL
	init();
    initSphereVBO(32, 32);
    initTextures();
    initShaders();

    // Zaregistrujeme callback funkcie
	glutDisplayFunc(render);                            
	glutReshapeFunc(reshape);                        
	glutKeyboardFunc(keyboard);                      
	glutSpecialFunc(special_keys);
    glutMouseFunc(mouse_klik);
    glutMotionFunc(mouse_move);
    glutTimerFunc(0, timer, 0);

    // Spustenie hlavneho okruhu zachytavania sprav
	glutMainLoop();

	return 0;
}
